// ==============================================================
//                   MagCompass MFD
//             Copyright (C) 2006 Chris Jeppesen
//                   All rights reserved
//
// MagCompass.cpp
// ==============================================================

#define STRICT
#define ORBITER_MODULE
#include "windows.h"
#include "orbitersdk.h"
#include "engunit.h"
#include "WMM2005.h"

#define RED RGB(255, 0, 0)
#define GREEN RGB(0, 255, 0)
#define YELLOW RGB(255, 255, 0)
#define DARK_YELLOW RGB(128, 128, 0)
#define WHITE RGB(255, 255, 255)
#define BLUE RGB(0, 0, 255)
#define GRAY RGB(160, 160, 160)
#define BRIGHTERGRAY RGB(200, 200, 200)

class MagCompass: public MFD {
public:
    VESSEL *vessel;
  MagCompass (DWORD w, DWORD h, VESSEL *v);
  static int MsgProc (UINT msg, UINT mfd, WPARAM wparam, LPARAM lparam);
  bool ConsumeKeyBuffered (DWORD key);
  ~MagCompass ();
  void Update (HDC hDC);
  void Print(HDC hDC, char* label, double value, int x, int l);

  int line( int );
  int X(double);

  int width,height,mid;
  VECTOR3 MaxA;
  double MaxAL;
  WMM2005 *W;
  HPEN whitePen;
  HPEN redPen;
};

// ==============================================================
// Global variables

int g_MFDmode;
MagCompass* thisAccelerometer;

// ==============================================================
// API interface

DLLCLBK void opcDLLInit (HINSTANCE hDLL) {
  static char *name = "MagCompass";   // MFD mode name
  MFDMODESPEC spec;
  spec.name = name;
  spec.key = OAPI_KEY_V;                // MFD mode selection key
  spec.msgproc = MagCompass::MsgProc;  // MFD mode callback function

  // Register the new MFD mode with Orbiter
  g_MFDmode = oapiRegisterMFDMode (spec);
  thisAccelerometer=NULL;

}

DLLCLBK void opcDLLExit (HINSTANCE hDLL) {
  // Unregister the custom MFD mode when the module is unloaded
  oapiUnregisterMFDMode (g_MFDmode);
}

// ==============================================================
// MFD class implementation

// Constructor
MagCompass::MagCompass (DWORD w, DWORD h, VESSEL *v): MFD (w, h, v) {
  vessel=v;
  // Add MFD initialisation here
  width=w;
  height=h;
  mid=w/2;
  MaxA=_V(0,0,0);
  MaxAL=0;
  W=new WMM2005();
  whitePen=CreatePen(PS_SOLID,1,WHITE);
  redPen=CreatePen(PS_SOLID,1,RED);
}

// Destructor
MagCompass::~MagCompass () {
  DeleteObject(redPen);
  DeleteObject(whitePen);
}

int MagCompass::line( int i ) {
  return (int)((float)i*((float)height/26.0));
}

void MagCompass::Print(HDC hDC, char* label, double value, int x, int l) {
  char buf[256];
  int len=sprintf(buf,label,value);
  TextOut(hDC,x,line(l),buf,len);
}

int MagCompass::X(double x) {
  return x*width;
}

// Repaint the MFD
void MagCompass::Update (HDC hDC) {
  char buf[256];
  int l=1;
  int nThrusters=vessel->GetThrusterCount();
  Title(hDC,"Magnetic Compass");
  double lat,lon,rad,bx,by,bz,ti,bh,dec,dip,truehead,maghead;
  vessel->GetEquPos(lon,lat,rad);
  oapiGetHeading(vessel->GetHandle(),&truehead);
  while(truehead<0) truehead+=2*PI;
  while(truehead>2*PI) truehead-=2*PI;
  double alt=(rad-6378137.0)/1000.0;
  W->geomg1(alt,lat,lon,2006,bx,by,bz,ti,bh,dec,dip);
  maghead=truehead-dec;
  while(maghead<0) maghead+=2*PI;
  while(maghead>2*PI) maghead-=2*PI;

  DisplayEngUnit(buf,"lat: %9.3f%c",lat*180/PI);
  TextOut(hDC,5,line(l++),buf,strlen(buf));
  DisplayEngUnit(buf,"lon: %9.3f%c",lon*180/PI);
  TextOut(hDC,5,line(l++),buf,strlen(buf));
  DisplayEngUnit(buf,"rad: %9.3f%cm",lat/1000);
  TextOut(hDC,5,line(l++),buf,strlen(buf));
  DisplayEngUnit(buf,"x:   %9.3f%cT",bx/1e9);
  TextOut(hDC,5,line(l++),buf,strlen(buf));
  DisplayEngUnit(buf,"y:   %9.3f%cT",by/1e9);
  TextOut(hDC,5,line(l++),buf,strlen(buf));
  DisplayEngUnit(buf,"z:   %9.3f%cT",bz/1e9);
  TextOut(hDC,5,line(l++),buf,strlen(buf));
  DisplayEngUnit(buf,"h:   %9.3f%cT",bh/1e9);
  TextOut(hDC,5,line(l++),buf,strlen(buf));
  DisplayEngUnit(buf,"ti:  %9.3f%cT",ti/1e9);
  TextOut(hDC,5,line(l++),buf,strlen(buf));
  DisplayEngUnit(buf,"dec: %9.3f%c",dec*180/PI);
  TextOut(hDC,5,line(l++),buf,strlen(buf));
  DisplayEngUnit(buf,"dip: %9.3f%c",dip*180/PI);
  TextOut(hDC,5,line(l++),buf,strlen(buf));

  DisplayEngUnit(buf,"THead: %9.3f%c",truehead*180/PI);
  TextOut(hDC,5,line(l++),buf,strlen(buf));
  DisplayEngUnit(buf,"MHead: %9.3f%c",maghead*180/PI);
  TextOut(hDC,5,line(l++),buf,strlen(buf));
  
  HGDIOBJ oldPen=SelectObject(hDC,whitePen);
  MoveToEx(hDC,X(0.5),X(0.5),NULL);
  LineTo(hDC,X(0.5+0.4*sin(-truehead)),X(0.5-0.4*cos(-truehead)));
  SelectObject(hDC,redPen);
  MoveToEx(hDC,X(0.5),X(0.5),NULL);
  LineTo(hDC,X(0.5+0.4*sin(-maghead)),X(0.5-0.4*cos(-maghead)));
  SelectObject(hDC,oldPen);
}

// MFD message parser
int MagCompass::MsgProc (UINT msg, UINT mfd, WPARAM wparam, LPARAM lparam) {
  switch (msg) {

  case OAPI_MSG_MFD_OPENED:
    // Our new MFD mode has been selected, so we create the MFD and
    // return a pointer to it.
    return (int)(new MagCompass (LOWORD(wparam), HIWORD(wparam), (VESSEL*)lparam));
  }
  return 0;
}


bool MagCompass::ConsumeKeyBuffered (DWORD key) {
  switch (key) {
    case OAPI_KEY_PERIOD:
      return true;
    case OAPI_KEY_COMMA:
      return true;
    case OAPI_KEY_LBRACKET:
      return true;
    case OAPI_KEY_RBRACKET:
      return true;
  }
  return false;
}


